#!/usr/bin/python
# -- Content-Encoding: UTF-8 --
"""

Discovery Provider API

:author: Scott Lewis
:copyright: Copyright 2020, Scott Lewis
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Scott Lewis

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import logging
from threading import RLock

from pelix.ipopo.decorators import (
    Provides,
    ComponentFactory,
    Instantiate,
    BindField,
    UnbindField,
    Requires,
)

from pelix.rsa import (
    get_string_plus_property,
    get_string_plus_property_value,
    ECF_ENDPOINT_CONTAINERID_NAMESPACE,
)
from pelix.rsa.endpointdescription import EndpointDescription

# ------------------------------------------------------------------------------
# Module version

__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# Standard logging
_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------


SERVICE_ENDPOINT_ADVERTISER = "pelix.rsa.discovery.endpointadvertiser"


class EndpointAdvertiser(object):
    """
    Endpoint advertiser service specification.  EndpointAdvertiser services
    are used to advertise exported remote services.  See EndpointAdvertiser
    class below.
    """

    def __init__(self):
        self._published_endpoints = {}
        self._published_endpoints_lock = RLock()

    def advertise_endpoint(self, endpoint_description):
        """
        Advertise and endpoint_description for remote discovery.
        If it hasn't already been a endpoint_description will be advertised via
        a some protocol.

        :param endpoint_description: an instance of EndpointDescription to
        advertise. Must not be None.
        :return: True if advertised, False if not
        (e.g. it's already been advertised)
        """
        endpoint_id = endpoint_description.get_id()
        with self._published_endpoints_lock:
            if self.get_advertised_endpoint(endpoint_id) is not None:
                return False

            advertise_result = self._advertise(endpoint_description)
            if advertise_result:
                self._add_advertised(endpoint_description, advertise_result)
                return True

            return False

    def update_endpoint(self, updated_ed):
        """
        Update a previously advertised endpoint_description.

        :param endpoint_description: an instance of EndpointDescription to
        update. Must not be None.
        :return: True if advertised, False if not
        (e.g. it's already been advertised)
        """
        endpoint_id = updated_ed.get_id()
        with self._published_endpoints_lock:
            if self.get_advertised_endpoint(endpoint_id) is None:
                return False

            advertise_result = self._update(updated_ed)
            if advertise_result:
                self._remove_advertised(endpoint_id)
                self._add_advertised(updated_ed, advertise_result)
                return True

            return False

    def unadvertise_endpoint(self, endpointid):
        """
        Unadvertise a previously-advertised endpointid (string).

        :param endpointid.  The string returned from ed.get_id() or
        the value of property endpoint.id.  Should not be None

        :return True if removed, False if not removed (hasn't been previously advertised
        by this advertiser
        """
        with self._published_endpoints_lock:
            with self._published_endpoints_lock:
                advertised = self.get_advertised_endpoint(endpointid)
                if not advertised:
                    return None

                unadvertise_result = self._unadvertise(advertised)
                if unadvertise_result:
                    self._remove_advertised(endpointid)

                return None

    def is_advertised(self, endpointid):
        """
        Is given endpointid been advertised by this advertiser.

        :param endpointid.  The string returned from ed.get_id() or the value
        of property endpoint.id.  Should not be None.
        :return True if the given endpointid has previously been
        advertised (and not unadvertised) by this advertiser
        """
        return self.get_advertised_endpoint(endpointid) is not None

    def get_advertised_endpoint(self, endpointid):
        """
        Get the advertised endpoint given endpointid.

        :param endpointid.  The string reeturned from ed.get_id() or the value
        of property endpoint.id.  Should not be None.

        :return tuple/2 with (endpoint_description,advertise_result), or
        None.
        """
        with self._published_endpoints_lock:
            return self._published_endpoints.get(endpointid, None)

    def get_advertised_endpoints(self):
        """
        Get all endpoints advertised by this advertiser.

        :returns dictionary of tuple/2 (endpoint_description,advertise_result).
        May return empty dictionary.
        """
        with self._published_endpoints_lock:
            return self._published_endpoints.copy()

    def _add_advertised(self, ed, advertise_result):
        with self._published_endpoints_lock:
            self._published_endpoints[ed.get_id()] = (ed, advertise_result)

    def _remove_advertised(self, endpointid):
        with self._published_endpoints_lock:
            return self._published_endpoints.pop(endpointid, None)

    def _advertise(self, endpoint_description):
        # pylint: disable=R0201, W0613
        raise Exception("Endpoint._advertise must be overridden by subclasses")

    def _update(self, endpoint_description):
        # pylint: disable=R0201, W0613
        raise Exception("Endpoint._update must be overridden by subclasses")

    def _unadvertise(self, advertised):
        # pylint: disable=R0201, W0613
        raise Exception(
            "Endpoint._unadvertise must be overridden by subclasses"
        )


# ------------------------------------------------------------------------------
# EndpointEvent implementation used to provide EndpointEventListener service
# instances with valid EndpointEvent by endpoint advertisers


class EndpointEvent(object):
    """
    EndpointEvents are used by endpoint advertisers to call
    EndpointEventListeners with the type of event (ADDED,REMOVED,MODIFIED)
    and the associated EndpointDescription.
    """

    ADDED = 1
    REMOVED = 2
    MODIFIED = 4

    def __init__(self, event_type, endpoint_description):
        assert event_type and isinstance(event_type, int)
        self._type = event_type
        assert endpoint_description and isinstance(
            endpoint_description, EndpointDescription
        )
        self._ed = endpoint_description

    def get_type(self):
        """
        Get the type of the EndpointEvent (ADDED|REMOVED|MODIFIED).

        :return event type (int)
        """
        return self._type

    def get_endpoint_description(self):
        """
        Get the EndpointDescription associated with this event.

        :return EndpointDescription instance associated with this event.
        """
        return self._ed

    def __str__(self):
        return "EndpointEvent(type={0},ed={1})".format(
            self.get_type(), self.get_endpoint_description()
        )


# ------------------------------------------------------------------------------
# Endpoint listener service specification
# This service specification is exposed by instances that wish to be
# notified by discovery providers when an EndpointEvent has occurred.
# For example, TopologyManagers will typically expose themselves as
# a service endpoint listener so that discovery subscribers can
# notify all such services when an endpoint event has been received.

SERVICE_ENDPOINT_LISTENER = "pelix.rsa.discovery.endpointeventlistener"
SERVICE_ENDPOINT_EVENT_LISTENER = SERVICE_ENDPOINT_LISTENER


class EndpointEventListener(object):
    # pylint: disable=R0903
    """
    Subclasses should override the endpoint_changed method
    so that they will receive notification (via an arbitrary
    thread) when an endpoint has been added, removed or modified
    """

    # Endpoint listener scope will be consulted when a
    # discovery provider receives an endpoint event.
    # The service that provides the SERVICE_ENDPOINT_LISTENER
    # specification should provided a ldap filter (or array of
    # filters) for filtering the endpoint description properties
    # against the filter for delivery.  For example, an implementer
    # of this service could define as one of its properties:
    # ENDPOINT_LISTENER_SCOPE: '(*=*)'
    # Such a filter will receive all endpoint notifications because
    # the filter '(*=*)' matches any set of endpoint description
    # properties.  If the filter matches the properties, then
    # the endpoint_changed method will be called
    ENDPOINT_LISTENER_SCOPE = "endpoint.listener.scope"

    def endpoint_changed(self, endpoint_event, matched_filter):
        # pylint: disable=W0613
        """
        Called by discovery providers when an endpoint has been
        ADDED,REMOVED or MODIFIED.

        :param endpoint_event an instance of EndpointEvent.  Will
        not be None.
        :param matched_filter the filter (as string) that matched
        this endpoint event listener service instance.
        """
        raise Exception("{0}.endpoint_changed not implemented".format(self))


@Requires("_event_listeners", SERVICE_ENDPOINT_LISTENER, True, True)
class EndpointSubscriber(object):
    # pylint: disable=R0903
    """
    Utility superclass for EndpointSubscribers.
    """

    def __init__(self):
        self._endpoint_event_listeners = []
        self._endpoint_event_listeners_lock = RLock()
        self._discovered_endpoints = {}
        self._discovered_endpoints_lock = RLock()

    @BindField("_event_listeners")
    def _add_endpoint_event_listener(self, field, listener, service_ref):
        # pylint: disable=W0613
        with self._endpoint_event_listeners_lock:
            self._endpoint_event_listeners.append((listener, service_ref))

    @UnbindField("_event_listeners")
    def _remove_endpoint_event_listener(self, field, listener, service_ref):
        # pylint: disable=W0613
        with self._endpoint_event_listeners_lock:
            try:
                return self._endpoint_event_listeners.remove(
                    (listener, service_ref)
                )
            except Exception:
                pass

    def _get_matching_endpoint_event_listeners(self, ed):
        result = []
        with self._discovered_endpoints_lock:
            ls = self._endpoint_event_listeners[:]
        for l in ls:
            svc_ref = l[1]
            filters = get_string_plus_property_value(
                svc_ref.get_property(
                    EndpointEventListener.ENDPOINT_LISTENER_SCOPE
                )
            )
            matching_filter = None
            if filters:
                for f in filters:
                    if ed.matches(f):
                        matching_filter = f
                        break
            if matching_filter:
                result.append((l[0], matching_filter))
        return result

    def _has_discovered_endpoint(self, ed_id):
        with self._discovered_endpoints_lock:
            ep = self._discovered_endpoints.get(ed_id, None)
            if ep:
                return ep[1]

            return None

    def _get_endpointids_for_sessionid(self, sessionid):
        result = []
        with self._discovered_endpoints_lock:
            for epid, ep in self._discovered_endpoints.items():
                if ep and sessionid == ep[0]:
                    result.append(epid)
        return result

    def _add_discovered_endpoint(self, sessionid, ed):
        with self._discovered_endpoints_lock:
            _logger.debug("_add_discovered_endpoint ed=%s", ed)
            self._discovered_endpoints[ed.get_id()] = (sessionid, ed)

    def _remove_discovered_endpoint(self, endpointid):
        with self._discovered_endpoints_lock:
            node = self._discovered_endpoints.pop(endpointid, None)
            if node:
                return node[1]

            return None

    def _fire_endpoint_event(self, event_type, ed):
        listeners = self._get_matching_endpoint_event_listeners(ed)
        if not listeners:
            logging.error(
                "EndpointSubscriber._fire_endpoint_event found no matching "
                "listeners for event_type=%s and endpoint=%s",
                event_type,
                ed,
            )
            return

        event = EndpointEvent(event_type, ed)
        for listener in listeners:
            try:
                listener[0].endpoint_changed(event, listener[1])
            except Exception:
                _logger.exception(
                    "Exception calling endpoint event "
                    "listener.endpoint_changed for listener=%s and event=%s",
                    listener,
                    event,
                )



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Pelix remote services: JSON-RPC implementation

Based on a modified version of the 3rd-party package jsonrpclib-pelix.

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import logging

# JSON-RPC module
import jsonrpclib.jsonrpc
from jsonrpclib.SimpleJSONRPCServer import SimpleJSONRPCDispatcher

# iPOPO decorators
from pelix.ipopo.decorators import (
    ComponentFactory,
    Requires,
    Validate,
    Invalidate,
    Property,
    Provides,
)

# Pelix constants
from pelix.utilities import to_str
import pelix.http
import pelix.remote
import pelix.remote.transport.commons as commons

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

JSONRPC_CONFIGURATION = "jsonrpc"
""" Remote Service configuration constant """

PROP_JSONRPC_URL = "{0}.url".format(JSONRPC_CONFIGURATION)
""" JSON-RPC servlet URL """

_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------


class _JsonRpcServlet(SimpleJSONRPCDispatcher):
    """
    A JSON-RPC servlet that can be registered in the Pelix HTTP service

    Calls the dispatch method given in the constructor
    """

    def __init__(self, dispatch_method, encoding=None):
        """
        Sets up the servlet
        """
        SimpleJSONRPCDispatcher.__init__(self, encoding=encoding)

        # Register the system.* functions
        self.register_introspection_functions()

        # Make a link to the dispatch method
        self._dispatch_method = dispatch_method

    def _simple_dispatch(self, name, params):
        """
        Dispatch method
        """
        try:
            # Internal method
            func = self.funcs[name]
        except KeyError:
            # Other method
            pass
        else:
            # Internal method found
            if isinstance(params, (list, tuple)):
                return func(*params)

            return func(**params)

        # Call the other method outside the except block, to avoid messy logs
        # in case of error
        return self._dispatch_method(name, params)

    def do_POST(self, request, response):
        # pylint: disable=C0103
        """
        Handles a HTTP POST request

        :param request: The HTTP request bean
        :param response: The HTTP response handler
        """
        try:
            # Get the request content
            data = to_str(request.read_data())

            # Dispatch
            result = self._marshaled_dispatch(data, self._simple_dispatch)

            # Send the result
            response.send_content(200, result, "application/json-rpc")
        except Exception as ex:
            response.send_content(
                500, "Internal error:\n{0}\n".format(ex), "text/plain"
            )


# ------------------------------------------------------------------------------


@ComponentFactory(pelix.remote.FACTORY_TRANSPORT_JSONRPC_EXPORTER)
@Provides(pelix.remote.SERVICE_EXPORT_PROVIDER)
@Requires("_http", pelix.http.HTTP_SERVICE)
@Property("_path", pelix.http.HTTP_SERVLET_PATH, "/JSON-RPC")
@Property(
    "_kinds",
    pelix.remote.PROP_REMOTE_CONFIGS_SUPPORTED,
    (JSONRPC_CONFIGURATION,),
)
class JsonRpcServiceExporter(commons.AbstractRpcServiceExporter):
    """
    JSON-RPC Remote Services exporter
    """

    def __init__(self):
        """
        Sets up the exporter
        """
        # Call parent
        super(JsonRpcServiceExporter, self).__init__()

        # Handled configurations
        self._kinds = None

        # HTTP Service
        self._http = None
        self._path = None

        # JSON-RPC servlet
        self._servlet = None

    def get_access(self):
        """
        Retrieves the URL to access this component
        """
        port = self._http.get_access()[1]
        return "http{2}://{{server}}:{0}{1}".format(
            port, self._path, "s" if self._http.is_https() else ""
        )

    def make_endpoint_properties(self, svc_ref, name, fw_uid):
        """
        Prepare properties for the ExportEndpoint to be created

        :param svc_ref: Service reference
        :param name: Endpoint name
        :param fw_uid: Framework UID
        :return: A dictionary of extra endpoint properties
        """
        return {PROP_JSONRPC_URL: self.get_access()}

    @Validate
    def validate(self, context):
        """
        Component validated
        """
        # Call parent
        super(JsonRpcServiceExporter, self).validate(context)

        # Create/register the servlet
        self._servlet = _JsonRpcServlet(self.dispatch)
        self._http.register_servlet(self._path, self._servlet)

    @Invalidate
    def invalidate(self, context):
        """
        Component invalidated
        """
        # Unregister the servlet
        self._http.unregister(None, self._servlet)

        # Call parent
        super(JsonRpcServiceExporter, self).invalidate(context)

        # Clean up members
        self._servlet = None


# ------------------------------------------------------------------------------


class _ServiceCallProxy(object):
    # pylint: disable=R0903
    """
    Service call proxy
    """

    def __init__(self, name, url):
        """
        Sets up the call proxy

        :param name: End point name
        :param url: End point URL
        """
        self.__name = name
        self.__url = url

    def __getattr__(self, name):
        """
        Prefixes the requested attribute name by the endpoint name
        """
        # Make a proxy for this call
        # This is an ugly trick to handle multi-threaded calls, as the
        # underlying proxy re-uses the same connection when possible: sometimes
        # it means sending a request before retrieving a result
        proxy = jsonrpclib.jsonrpc.ServerProxy(self.__url)
        return getattr(proxy, "{0}.{1}".format(self.__name, name))


@ComponentFactory(pelix.remote.FACTORY_TRANSPORT_JSONRPC_IMPORTER)
@Provides(pelix.remote.SERVICE_IMPORT_ENDPOINT_LISTENER)
@Property(
    "_kinds",
    pelix.remote.PROP_REMOTE_CONFIGS_SUPPORTED,
    (JSONRPC_CONFIGURATION,),
)
class JsonRpcServiceImporter(commons.AbstractRpcServiceImporter):
    """
    JSON-RPC Remote Services importer
    """

    def __init__(self):
        """
        Sets up the exporter
        """
        # Call parent
        super(JsonRpcServiceImporter, self).__init__()

        # Component properties
        self._kinds = None

    def make_service_proxy(self, endpoint):
        """
        Creates the proxy for the given ImportEndpoint

        :param endpoint: An ImportEndpoint bean
        :return: A service proxy
        """
        # Get the access URL
        access_url = endpoint.properties.get(PROP_JSONRPC_URL)
        if not access_url:
            # No URL information
            _logger.warning("No access URL given: %s", endpoint)
            return None

        if endpoint.server is not None:
            # Server information given
            access_url = access_url.format(server=endpoint.server)
        else:
            # Use the local IP as the source server, just in case
            local_server = "localhost"
            access_url = access_url.format(server=local_server)

        # Return the proxy
        return _ServiceCallProxy(endpoint.name, access_url)

    def clear_service_proxy(self, endpoint):
        """
        Destroys the proxy made for the given ImportEndpoint

        :param endpoint: An ImportEndpoint bean
        """
        # Nothing to do
        return



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
iPOPO shell commands

Provides commands to the Pelix shell to get the state of iPOPO instances.

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import logging

# Pelix
from pelix.ipopo.decorators import (
    ComponentFactory,
    Requires,
    Provides,
    Instantiate,
)
import pelix.ipopo.constants
import pelix.shell

from pelix.shell.completion.decorators import (
    Completion,
    FACTORY,
    COMPONENT,
    DUMMY,
    FACTORY_PROPERTY,
)

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------


def ipopo_state_to_str(state):
    """
    Converts the state of a component instance to its string representation

    :param state: The state of an iPOPO component
    :return: A string representation of the state
    """
    ipopo_states = {
        0: "INVALID",
        1: "VALID",
        2: "KILLED",
        3: "VALIDATING",
        4: "ERRONEOUS",
    }

    return ipopo_states.get(state, "Unknown state ({0})".format(state))


# ------------------------------------------------------------------------------


@ComponentFactory("ipopo-shell-commands-factory")
@Requires("_ipopo", pelix.ipopo.constants.SERVICE_IPOPO)
@Requires("_utils", pelix.shell.SERVICE_SHELL_UTILS)
@Provides(pelix.shell.SERVICE_SHELL_COMMAND)
@Instantiate("ipopo-shell-commands")
class IPopoCommands(object):
    """
    iPOPO shell commands
    """

    def __init__(self):
        """
        Sets up the object
        """
        self._ipopo = None
        self._utils = None

    @staticmethod
    def get_namespace():
        """
        Retrieves the name space of this command handler
        """
        return "ipopo"

    def get_methods(self):
        """
        Retrieves the list of tuples (command, method) for this command handler
        """
        return [
            ("factories", self.list_factories),
            ("factory", self.factory_details),
            ("instances", self.list_instances),
            ("waiting", self.list_waitings),
            ("instance", self.instance_details),
            ("instantiate", self.instantiate),
            ("kill", self.kill),
            ("retry", self.retry_erroneous),
        ]

    def list_factories(self, session, name=None):
        """
        Lists the available iPOPO component factories
        """
        header = ("Factory", "Bundle")

        factories = self._ipopo.get_factories()
        if name is not None:
            # Filter factories by name
            factories = [factory for factory in factories if name in factory]

        lines = sorted(
            (name, self._ipopo.get_factory_bundle(name)) for name in factories
        )

        session.write(self._utils.make_table(header, lines))
        if name is None:
            session.write_line("{0} factories available", len(lines))
        else:
            session.write_line("{0} filtered factories", len(lines))

    def list_instances(self, session, name=None):
        """
        Lists the active iPOPO component instances
        """
        headers = ("Name", "Factory", "State")

        instances = self._ipopo.get_instances()
        if name is not None:
            # Filter instances by name
            instances = [
                instance for instance in instances if name in instance[0]
            ]

        # Lines are already sorted
        lines = (
            (name, factory, ipopo_state_to_str(state))
            for name, factory, state in instances
        )

        session.write(self._utils.make_table(headers, lines))
        if name is None:
            session.write_line("{0} components running", len(instances))
        else:
            session.write_line("{0} filtered components", len(instances))

    def list_waitings(self, session, name=None):
        """
        Lists the components waiting to be instantiated
        """
        headers = ("Name", "Factory", "Missing handlers")

        components = self._ipopo.get_waiting_components()
        if name is not None:
            # Filter components by name
            components = [
                component for component in components if name in component[0]
            ]

        # Lines are already sorted
        lines = (
            (name, factory, ", ".join(missing))
            for name, factory, missing in components
        )

        session.write(self._utils.make_table(headers, lines))
        if name is None:
            session.write_line(
                "{0} components in the waiting queue", len(components)
            )
        else:
            session.write_line("{0} filtered components", len(components))

    @Completion(FACTORY)
    def factory_details(self, session, name):
        """
        Prints the details of the given component factory
        """
        try:
            details = self._ipopo.get_factory_details(name)
        except ValueError as ex:
            session.write_line(
                "Error getting details about '{0}': {1}", name, ex
            )
            return False

        lines = [
            "Name  : {0}".format(details["name"]),
            "Bundle: {0}".format(details["bundle"]),
        ]

        properties = details.get("properties", None)
        if properties:
            lines.append("Properties:")
            prop_headers = ("Key", "Default value")
            prop_lines = [
                (str(key), str(value)) for key, value in properties.items()
            ]
            lines.append(self._utils.make_table(prop_headers, prop_lines))

        services = details.get("services", None)
        if services:
            lines.append("Provided services:")
            lines.extend("\t{0}".format(spec) for spec in services)
            lines.append("")

        requirements = details.get("requirements", None)
        if requirements:
            lines.append("Requirements:")
            req_headers = (
                "ID",
                "Specification",
                "Filter",
                "Aggregate",
                "Optional",
            )
            req_lines = [
                (
                    item["id"],
                    item["specification"],
                    item["filter"],
                    item["aggregate"],
                    item["optional"],
                )
                for item in requirements
            ]

            lines.append(self._utils.make_table(req_headers, req_lines, "\t"))

        handlers = details.get("handlers", None)
        if handlers:
            lines.append("Handlers:")
            handlers_headers = ("ID", "Configuration")
            handlers_lines = [(key, handlers[key]) for key in sorted(handlers)]
            lines.append(
                self._utils.make_table(handlers_headers, handlers_lines, "\t")
            )

        session.write("\n".join(lines))
        return None

    @Completion(COMPONENT)
    def instance_details(self, session, name):
        """
        Prints the details of the given component instance
        """
        try:
            details = self._ipopo.get_instance_details(name)
        except ValueError as ex:
            session.write_line(
                "Error getting details about '{0}': {1}", name, ex
            )
            return False

        # Basic information
        lines = [
            "Name.....: {0}".format(details["name"]),
            "Factory..: {0}".format(details["factory"]),
            "Bundle ID: {0}".format(details["bundle_id"]),
            "State....: {0}".format(ipopo_state_to_str(details["state"])),
            "Services.:",
        ]

        # Provided services
        lines.extend(
            "\t{0}".format(svc_reference)
            for svc_reference in details["services"].values()
        )

        # Requirements
        lines.append("Dependencies:")
        for field, infos in details["dependencies"].items():
            lines.append("\tField: {0}".format(field))
            lines.append(
                "\t\tSpecification: {0}".format(infos["specification"])
            )
            lines.append("\t\tFilter......: {0}".format(infos["filter"]))
            lines.append("\t\tOptional.....: {0}".format(infos["optional"]))
            lines.append("\t\tAggregate....: {0}".format(infos["aggregate"]))

            lines.append("\t\tHandler......: {0}".format(infos["handler"]))
            lines.append("\t\tBindings:")
            for ref in infos["bindings"]:
                lines.append("\t\t\t{0}".format(ref))

        # Properties
        lines.append("Properties:")
        lines.append(
            self._utils.make_table(
                ("Key", "Value"), sorted(details["properties"].items()), "\t"
            )
        )

        # Error trace, for erroneous components
        error_trace = details["error_trace"]
        if error_trace:
            lines.append("Error trace:")
            lines.append(error_trace)

        lines.append("")
        session.write("\n".join(lines))
        return None

    @Completion(FACTORY, DUMMY, FACTORY_PROPERTY, multiple=True)
    def instantiate(self, session, factory, name, **properties):
        """
        Instantiates a component of the given factory with the given name and
        properties
        """
        try:
            self._ipopo.instantiate(factory, name, properties)
            session.write_line("Component '{0}' instantiated.", name)
            # Return the instance name as a result
            return name
        except ValueError as ex:
            session.write_line("Invalid parameter: {0}", ex)
        except TypeError as ex:
            session.write_line("Invalid factory: {0}", ex)
        except Exception as ex:
            session.write_line("Error instantiating the component: {0}", ex)
            _logger.exception("Error instantiating the component")

        # We're here if an exception occurred
        return False

    @Completion(COMPONENT)
    def kill(self, session, name):
        """
        Kills the given component instance
        """
        try:
            self._ipopo.kill(name)
            session.write_line("Component '{0}' killed.", name)
        except ValueError as ex:
            session.write_line("Invalid parameter: {0}", ex)
            return False

        return None

    @Completion(COMPONENT, FACTORY_PROPERTY, multiple=True)
    def retry_erroneous(self, session, name, **properties):
        """
        Removes the erroneous flag from a component and retries to validate it
        """
        try:
            new_state = self._ipopo.retry_erroneous(name, properties)
            session.write_line(
                "Component '{0}' is now in state {1}.",
                name,
                ipopo_state_to_str(new_state),
            )
        except ValueError as ex:
            session.write_line("Invalid parameter: {0}", ex)
            return False

        return None



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
XMPP shell: XMPP interface for the Pelix shell

This module depends on the sleekxmpp package: http://sleekxmpp.com/

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import argparse
import collections
import logging
import sys

try:
    # Python 2
    from StringIO import StringIO
except ImportError:
    # Python 3
    from io import StringIO

# SleekXMPP
import sleekxmpp

# iPOPO decorators
from pelix.ipopo.decorators import (
    ComponentFactory,
    Requires,
    Property,
    Validate,
    Invalidate,
    HiddenProperty,
)
from pelix.ipopo.constants import use_ipopo

# Pelix utilities
import pelix.framework
import pelix.misc.xmpp
import pelix.threadpool
import pelix.utilities

# Shell constants
from pelix.shell.console import (
    make_common_parser,
    handle_common_arguments,
    remove_duplicates,
)
import pelix.shell
import pelix.shell.beans as beans

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

# Logger
_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------


class _XmppOutStream(object):
    """
    File-like XMPP output. For shell IOHandler use only
    """

    def __init__(self, client, target):
        """
        Sets up the stream

        :param client: XMPP client
        :param target: Output target JID
        """
        self._client = client
        self._target = target
        self._buffer = StringIO()

        # Indicate to the I/O handler that we want strings, not bytes
        self.encoding = "utf-8"

    def write(self, data):
        """
        Writes data to a buffer
        """
        self._buffer.write(data)

    def flush(self):
        """
        Sends buffered data to the target
        """
        # Flush buffer
        content = self._buffer.getvalue()
        self._buffer = StringIO()

        if content:
            # Send message
            self._client.send_message(self._target, content, mtype="chat")


class _XmppInStream(object):
    """
    File-like XMPP input. For shell IOHandler use only
    """

    def __init__(self, xmpp_ui, source_jid):
        """
        Sets up the object

        :param xmpp_ui: The IPopoXMPPShell object
        """
        self._ui = xmpp_ui
        self._jid = source_jid

    def readline(self):
        """
        Waits for a line from the XMPP client
        """
        # Wait for content from the user
        return self._ui.read_from(self._jid)


# ------------------------------------------------------------------------------


@ComponentFactory(pelix.shell.FACTORY_XMPP_SHELL)
@Requires("_shell", pelix.shell.SERVICE_SHELL)
@Property("_host", "shell.xmpp.server", "localhost")
@Property("_port", "shell.xmpp.port", 5222)
@Property("_jid", "shell.xmpp.jid")
@HiddenProperty("_password", "shell.xmpp.password")
@Property("_use_tls", "shell.xmpp.tls", "1")
@Property("_use_ssl", "shell.xmpp.ssl", "0")
class IPopoXMPPShell(object):
    """
    The iPOPO XMPP Shell, based on the Pelix Shell
    """

    def __init__(self):
        """
        Sets up the component
        """
        # Injected fields
        self._shell = None
        self._host = None
        self._port = 5222
        self._jid = None
        self._password = None
        self._use_tls = True
        self._use_ssl = False

        # XMPP Bot
        self.__bot = None

        # Shell sessions: JID -> ShellSession
        self.__sessions = {}

        # Waiting for a message from the given JID
        self.__waiting = {}

        # Task queue thread
        self.__pool = pelix.threadpool.ThreadPool(1, logname="XMPPShell")

    @staticmethod
    def __normalize_int(value, default=0):
        """
        Normalizes an integer value
        """
        try:
            return int(value)
        except (ValueError, TypeError):
            return default

    @Validate
    def validate(self, _):
        """
        Component validated
        """
        # Normalize values
        self._port = self.__normalize_int(self._port, 5222)
        self._use_tls = bool(self.__normalize_int(self._use_tls, 1))
        self._use_ssl = bool(self.__normalize_int(self._use_ssl, 0))

        # Start the thread pool
        self.__pool.start()

        _logger.info(
            """XMPP shell:
- JID: %s
- host: %s
- port: %s""",
            self._jid,
            self._host,
            self._port,
        )

        # Create the bot. Negative priority avoids listening to human messages
        self.__bot = pelix.misc.xmpp.BasicBot(self._jid, self._password)
        self.__bot.auto_authorize = True

        # Register to events
        self.__bot.add_event_handler("message", self.__on_message)
        self.__bot.add_event_handler("session_start", self.__on_start)
        self.__bot.add_event_handler("session_end", self.__on_end)
        self.__bot.add_event_handler("socket_error", self.__on_error)
        self.__bot.add_event_handler("got_offline", self.__on_offline)

        # Connect to the server
        self.__bot.connect(
            self._host, self._port, False, self._use_tls, self._use_ssl
        )

    @Invalidate
    def invalidate(self, _):
        """
        Component invalidated
        """
        # Stop the thread pool
        self.__pool.stop()

        # Disconnect the bot
        self.__bot.disconnect()
        self.__bot = None

        # Clean up
        self.__sessions.clear()

    @staticmethod
    def __on_error(error):
        """
        A socket error occurred

        :param error: The socket error
        """
        _logger.exception("A socket error occurred: %s", error)

    def __on_start(self, _):
        """
        XMPP session started
        """
        _logger.info(
            "XMPP shell connected with JID: %s", self.__bot.boundjid.full
        )

    def __on_end(self, _):
        """
        XMPP session ended
        """
        _logger.info(
            "XMPP shell disconnected from %s", self.__bot.boundjid.full
        )

    def __on_offline(self, data):
        """
        XMPP client got offline
        :param data: Message stanza
        """
        source_jid = data["from"].full
        try:
            # Unsubscribe to presence events
            self.__bot.sendPresence(pto=source_jid, ptype="unsubscribe")

            # Delete the corresponding session
            del self.__sessions[source_jid]
        except KeyError:
            # Unknown JID
            pass

    def __on_message(self, data):
        """
        Got an XMPP message

        :param data: Message stanza (see SleekXMPP)
        """
        if data["type"] in ("normal", "chat"):
            # Got a message
            body = data["body"].strip()
            if body:
                # Valid content, check the sender
                sender = data["from"].full
                try:
                    # Are we listening to this JID ?
                    event = self.__waiting[sender].popleft()
                except (KeyError, IndexError):
                    # Not waiting for a message from this JID,
                    # treat the message in a the task thread
                    self.__pool.enqueue(self.handle_message, sender, body)
                else:
                    # Set the event, with the content of the message as data
                    event.set(body)

    def handle_message(self, source_jid, content):
        """
        Handles an XMPP message and reply to the source

        :param source_jid: JID of the message sender
        :param content: Content of the message
        """
        try:
            # Use the existing session
            session = self.__sessions[source_jid]
        except KeyError:
            # Subscribe to presence messages
            self.__bot.sendPresence(pto=source_jid, ptype="subscribe")

            # Create and store the session
            session = self.__sessions[source_jid] = beans.ShellSession(
                beans.IOHandler(
                    _XmppInStream(self, source_jid),
                    _XmppOutStream(self.__bot, source_jid),
                ),
                {"xmpp.jid": source_jid},
            )

        self._shell.execute(content, session)

    def read_from(self, jid):
        """
        Returns the next message read from the given JID

        :param jid: Source JID
        :return: The next content send by this JID
        """
        # Prepare an event
        event = pelix.utilities.EventData()
        self.__waiting.setdefault(jid, collections.deque()).append(event)

        # Wait for the event to be set...
        event.wait()

        # Return the message content
        return event.data


# ------------------------------------------------------------------------------


def main(argv=None):
    """
    Entry point

    :param argv: Script arguments (None for sys.argv)
    :return: An exit code or None
    """
    # Prepare arguments
    parser = argparse.ArgumentParser(
        prog="pelix.shell.xmpp",
        parents=[make_common_parser()],
        description="Pelix XMPP Shell",
    )

    group = parser.add_argument_group("XMPP options")
    group.add_argument("-j", "--jid", dest="jid", help="Jabber ID")
    group.add_argument("--password", dest="password", help="JID password")
    group.add_argument("-s", "--server", dest="server", help="XMPP server host")
    group.add_argument(
        "-p",
        "--port",
        dest="port",
        type=int,
        default=5222,
        help="XMPP server port",
    )
    group.add_argument(
        "--tls",
        dest="use_tls",
        action="store_true",
        help="Use a STARTTLS connection",
    )
    group.add_argument(
        "--ssl",
        dest="use_ssl",
        action="store_true",
        help="Use an SSL connection",
    )

    # Parse them
    args = parser.parse_args(argv)

    # Handle common arguments
    init = handle_common_arguments(args)

    # Quiet down the SleekXMPP logger
    if not args.verbose:
        logging.getLogger("sleekxmpp").setLevel(logging.WARNING)

    if not args.server and not args.jid:
        _logger.error("No JID nor server given. Abandon.")
        sys.exit(1)

    # Get the password if necessary
    password = args.password
    if args.jid and args.password is None:
        try:
            import getpass
        except ImportError:
            _logger.error(
                "getpass() unavailable: give a password in command line"
            )
        else:
            try:
                password = getpass.getpass()
            except getpass.GetPassWarning:
                pass

    # Get the server from the JID, if necessary
    server = args.server
    if not server:
        server = sleekxmpp.JID(args.jid).domain

    # Set the initial bundles
    bundles = [
        "pelix.ipopo.core",
        "pelix.shell.core",
        "pelix.shell.ipopo",
        "pelix.shell.console",
        "pelix.shell.xmpp",
    ]
    bundles.extend(init.bundles)

    # Use the utility method to create, run and delete the framework
    framework = pelix.framework.create_framework(
        remove_duplicates(bundles), init.properties
    )
    framework.start()

    # Instantiate a Remote Shell
    with use_ipopo(framework.get_bundle_context()) as ipopo:
        ipopo.instantiate(
            pelix.shell.FACTORY_XMPP_SHELL,
            "xmpp-shell",
            {
                "shell.xmpp.server": server,
                "shell.xmpp.port": args.port,
                "shell.xmpp.jid": args.jid,
                "shell.xmpp.password": password,
                "shell.xmpp.tls": args.use_tls,
                "shell.xmpp.ssl": args.use_ssl,
            },
        )

    # Instantiate configured components
    init.instantiate_components(framework.get_bundle_context())

    try:
        framework.wait_for_stop()
    except KeyboardInterrupt:
        framework.stop()


if __name__ == "__main__":
    # Run the entry point
    sys.exit(main() or 0)



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Tests the utility module

:author: Thomas Calmant
"""

# Same version as the tested bundle
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

# Standard library
import random
import sys
import threading
import time

try:
    import unittest2 as unittest
except ImportError:
    import unittest

# Pelix
import pelix.constants
import pelix.framework
import pelix.utilities as utilities

# Tests
from tests.interfaces import IEchoService

# ------------------------------------------------------------------------------


class SynchronizationUtilitiesTest(unittest.TestCase):
    """
    Tests for utility module synchronization methods
    """
    def setUp(self):
        """
        Sets up the test
        """
        self.lock = threading.Lock()

    def testIsLock(self):
        """
        Tests the is_lock method
        """
        valid = (threading.Lock(), threading.RLock(), threading.Semaphore(),
                 threading.Condition())
        invalid = (None, "", 1234, object())

        for test in valid:
            self.assertTrue(utilities.is_lock(test),
                            "Valid lock not detected: {0}"
                            .format(type(test).__name__))

        for test in invalid:
            self.assertFalse(utilities.is_lock(test),
                             "Invalid lock not detected: {0}"
                             .format(type(test).__name__))

    @utilities.SynchronizedClassMethod('lock')
    def testSynchronizedClassMethod(self):
        """
        Tests the @SynchronizedClassMethod decorator
        """
        # Just test if the lock is really locked
        self.assertFalse(self.lock.acquire(False), "Method is not locked")

    def testSynchronizedMethod(self, no_lock=False):
        """
        Tests the @Synchronized decorator, with or without a given lock

        :param no_lock: If True, create the lock, else let the decorator do it
        """
        # Thread results: ID -> starting time
        result = {}

        # Synchronization lock
        if no_lock:
            lock = None
        else:
            lock = threading.Lock()

        @utilities.Synchronized(lock)
        def sleeper(wait, sleep_id):
            """
            Sleeps during *wait* seconds
            """
            if lock is not None:
                self.assertFalse(lock.acquire(False), "Lock not locked")

            result[sleep_id] = time.time()
            time.sleep(wait)

        # Get first call time
        start = time.time()

        # Launch first waiter
        thread1 = threading.Thread(target=sleeper, args=(.5, 1))
        thread1.start()

        # Wait a little before starting 2nd thread: on Windows, thread 2
        # can start before thread 1
        time.sleep(.1)

        # Launch second waiter
        thread2 = threading.Thread(target=sleeper, args=(0, 2))
        thread2.start()

        # Get intermediate time
        interm = time.time()

        # Wait for threads
        for thread in (thread1, thread2):
            thread.join()

        # Validate conditions :
        # .. Thread 1 started after start (obvious)
        self.assertGreaterEqual(result[1], start, "Thread 1 started too soon")

        # .. Thread 2 started at least 0.4 secs after thread 1
        # (due to the lock)
        # (0.4 instead of 0.5: some systems are not that precise)
        self.assertGreaterEqual(result[2], result[1] + .4,
                                "Thread 2 started too soon (after {0}s)"
                                .format(result[2] - result[1]))

        # .. Thread 2 must not have blocked the main thread
        self.assertGreater(result[2], interm,
                           "Thread 2 blocked the main thread")

    def testSynchronizedMethod2(self):
        """
        Tests the @Synchronized decorator, without a given lock
        """
        self.testSynchronizedMethod(True)

    def testNoLockException(self):
        """
        Verifies that @SynchronizedClassMethod raises an error when no lock
        is given
        """
        try:
            @utilities.SynchronizedClassMethod()
            def dummy():
                pass

            self.fail("@SynchronizedClassMethod() should raise a ValueError")
        except ValueError:
            # We must be there to succeed
            pass

        try:
            @utilities.SynchronizedClassMethod(None)
            def dummy():
                pass

            self.fail("@SynchronizedClassMethod(None) should raise a "
                      "ValueError")
        except ValueError:
            # We must be there to succeed
            pass

    def testNoneLockException(self):
        """
        Verifies that @SynchronizedClassMethod raises an error when a None lock
        is used for locking
        """
        self.lock = None
        self.assertRaises(AttributeError, self.testSynchronizedClassMethod)

# ------------------------------------------------------------------------------


class UtilitiesTest(unittest.TestCase):
    """
    Tests for utility module methods
    """
    def testReadOnlyProperty(self):
        """
        Tests the read only property generator
        """
        value_1 = 42
        value_2 = random.random()

        # Prepare the class members
        class Dummy(object):
            inside = utilities.read_only_property(value_1)

        Dummy.outside = utilities.read_only_property(value_2)

        # Work on an instance
        instance = Dummy()

        # Test read values
        self.assertEqual(instance.inside, value_1,
                         "Invalid initial value (in)")
        self.assertEqual(instance.outside, value_2,
                         "Invalid initial value (out)")

        # Test set values
        try:
            instance.inside = random.random()
            self.fail("Instance value (in) must not be modified.")

        except AttributeError:
            # We must be there
            pass

        try:
            instance.outside = random.random()
            self.fail("Instance value (out) must not be modified.")

        except AttributeError:
            # We must be there
            pass

        # Test final values (just in case)
        self.assertEqual(instance.inside, value_1,
                         "Invalid final value (in)")
        self.assertEqual(instance.outside, value_2,
                         "Invalid final value (out)")

    def testRemoveAllOccurrences(self):
        """
        Tests the remove_all_occurrences() method
        """
        try:
            # Must not raise an exception
            utilities.remove_all_occurrences(None, 12)
        except:
            self.fail("remove_all_occurrences(None) must not raise an "
                      "exception")

        min_value = -1
        max_value = 4

        # Create a random list
        list_org = []
        for i in range(0, random.randint(10, 20)):
            list_org.append(random.randint(min_value, max_value))

        # Create a copy
        list_copy = list_org[:]

        # Pick a random element
        for i in range(min_value, max_value + 1):
            # Get the original count
            count_base = list_org.count(i)
            self.assertEqual(list_copy.count(i), count_base,
                             "Copies doesn't have the same count of values")

            # Get the current length of the copy
            len_base = len(list_copy)

            # Remove the element
            utilities.remove_all_occurrences(list_copy, i)

            # The new count must be 0
            self.assertEqual(list_copy.count(i), 0, "Some references remain")

            # The new length must be len_base - count_base
            self.assertEqual(len(list_copy), len_base - count_base,
                             "Incorrect new list size")

    def testIsString(self):
        """
        Tests the is_string() method
        """
        valid = ["", "aaa", str(42)]
        invalid = [42, None, [], {}, tuple()]

        if sys.version_info[0] >= 3:
            # Python 3: test bytes
            invalid.extend((b"", b"aaa"))
        else:
            # Python 2: test unicode
            valid.extend((unicode(""), unicode("aaa"), unicode(42)))

        for value in valid:
            self.assertTrue(utilities.is_string(value),
                            "'{0}' is a string".format(value))

        for value in invalid:
            self.assertFalse(utilities.is_string(value),
                             "'{0}' is not a string".format(value))

    def testAddRemoveListener(self):
        """
        Tests add/remove listener methods
        """
        registry = []
        values = (42, "test", (1, 2, 3))

        # None value
        self.assertFalse(utilities.add_listener(registry, None),
                         "None value must not be accepted")

        self.assertFalse(utilities.remove_listener(registry, None),
                         "None value must not be accepted")

        for value in values:
            # Non-present value
            self.assertFalse(utilities.remove_listener(registry, value),
                             "Non-present value removed")

            # Add value
            self.assertTrue(utilities.add_listener(registry, value),
                            "Value has been refused")
            self.assertEqual(registry.count(value), 1,
                             "Value not inserted in registry")

            # Second add
            self.assertFalse(utilities.add_listener(registry, value),
                             "Value has been added twice")
            self.assertEqual(registry.count(value), 1,
                             "Value has been added twice")

        for value in values:
            # Remove value
            self.assertTrue(utilities.remove_listener(registry, value),
                            "Value has not been removed")
            # Ensure the value has been remove
            self.assertEqual(registry.count(value), 0,
                             "Value has not been removed")

            # Second removal
            self.assertFalse(utilities.remove_listener(registry, value),
                             "Value has been removed twice")

    def testUseService(self):
        """
        Tests utilities.use_service()
        """
        framework = pelix.framework.create_framework([])
        framework.start()
        context = framework.get_bundle_context()

        # Try without the service reference: TypeError
        self.assertRaises(TypeError,
                          utilities.use_service(context, None).__enter__)

        # Start the service bundle
        bundle = context.install_bundle("tests.framework.service_bundle")
        bundle.start()

        # Get the service reference
        svc_ref = context.get_service_reference(IEchoService)

        # Use it
        with utilities.use_service(context, svc_ref) as service:
            # Test the usage information
            self.assertIn(context.get_bundle(),
                          svc_ref.get_using_bundles(),
                          "Bundles using the service not updated")

            # Get the service the Pelix way
            got_service = context.get_service(svc_ref)

            # Test the service object
            self.assertIs(service, got_service, "Found a different service.")

            # Clean up the test usage
            context.unget_service(svc_ref)
            got_service = None

            # Re-test the usage information
            self.assertIn(context.get_bundle(),
                          svc_ref.get_using_bundles(),
                          "Bundles using service not kept")

        # Test the usage information
        self.assertNotIn(context.get_bundle(),
                         svc_ref.get_using_bundles(),
                         "Bundles using service kept after block")

        # Stop the iPOPO bundle
        bundle.stop()

        # Ensure the service is not accessible anymore
        self.assertRaises(pelix.constants.BundleException,
                          utilities.use_service(context, svc_ref).__enter__)

        # Uninstall the bundle
        bundle.uninstall()

        # Ensure the service is not accessible anymore
        self.assertRaises(pelix.constants.BundleException,
                          utilities.use_service(context, svc_ref).__enter__)

    def testToIterable(self):
        """
        Tests the to_iterable() method
        """
        # None value
        self.assertIsNone(utilities.to_iterable(None, True),
                          "None value refused")
        self.assertListEqual(utilities.to_iterable(None, False), [],
                             "None value accepted")

        # Check iterable types
        for clazz in (list, tuple, set, frozenset):
            iterable = clazz()
            self.assertIs(utilities.to_iterable(iterable), iterable,
                          "to_iterable() didn't returned the original object")

        # Check other types
        for value in ("hello", 123, {1: 2}, object()):
            self.assertListEqual(utilities.to_iterable(value), [value],
                                 "to_iterable() didn't returned a list")

# ------------------------------------------------------------------------------


class CountdownEventTest(unittest.TestCase):
    """
    Tests for the CountdownEvent class
    """
    def testInitCheck(self):
        """
        Tests the value check when creating the event
        """
        for invalid in (-1, 0):
            self.assertRaises(ValueError, utilities.CountdownEvent, invalid)

    def testSteps(self):
        """
        Tests the count down event behavior
        """
        event = utilities.CountdownEvent(3)
        # Stepping...
        self.assertFalse(event.step(), "Finished on first step...")
        self.assertFalse(event.is_set(), "Set on first step...")
        self.assertFalse(event.step(), "Finished on second step...")
        self.assertFalse(event.is_set(), "Set on second step...")

        # Last one
        self.assertTrue(event.step(), "Not done on last step...")
        self.assertTrue(event.is_set(), "Not set on last step...")

        # No more
        self.assertRaises(ValueError, event.step)
        self.assertTrue(event.is_set(), "Not set after last step...")

    def testWait(self):
        """
        Tests the wait() method
        """
        event = utilities.CountdownEvent(1)
        self.assertFalse(event.wait(.1), "Timed out wait must return False")

        start = time.time()
        threading.Timer(1, event.step).start()
        self.assertFalse(event.wait(.1), "Timed out wait must return False")
        self.assertTrue(event.wait(), "Wait should return true on set")
        self.assertLessEqual(time.time() - start, 2, "Too long to wait")

        self.assertTrue(event.wait(.5),
                        "Already set event shoudn't block wait()")
        self.assertTrue(event.wait(),
                        "Already set event shoudn't block wait()")

# ------------------------------------------------------------------------------


class EventDataTest(unittest.TestCase):
    """
    Tests for the EventData class
    """
    def testSetClear(self):
        """
        Tests set() and clear() operations
        """
        # Initial condition
        event = utilities.EventData()
        self.assertFalse(event.is_set(), "Event initially set")
        self.assertIsNone(event.data, "Non-None data")
        self.assertIsNone(event.exception, "Non-None exception")

        # No-data set
        event.set()
        self.assertTrue(event.is_set(), "Event not set")
        self.assertIsNone(event.data, "Non-None data")
        self.assertIsNone(event.exception, "Non-None exception")

        # Clear
        event.clear()
        self.assertFalse(event.is_set(), "Event still set")
        self.assertIsNone(event.data, "Non-None data")
        self.assertIsNone(event.exception, "Non-None exception")

        # Set data
        data = object()
        event.set(data)
        self.assertTrue(event.is_set(), "Event not set")
        self.assertIs(event.data, data, "Invalid event data")
        self.assertIsNone(event.exception, "Non-None exception")

        # Clear
        event.clear()
        self.assertFalse(event.is_set(), "Event still set")
        self.assertIsNone(event.data, "Non-None data")
        self.assertIsNone(event.exception, "Non-None exception")

    def testException(self):
        """
        Tests the exception storage
        """
        event = utilities.EventData()

        # "Raise" an exception
        exception = Exception("Some dummy exception")
        event.raise_exception(exception)

        # Check content
        self.assertTrue(event.is_set(), "Event has not been set")
        self.assertIsNone(event.data, "Non-None data")
        self.assertIs(event.exception, exception, "Invalid exception")

        # Check the behavior of "wait"
        try:
            event.wait()
        except Exception as ex:
            self.assertIs(ex, exception, "Not the same exception")
            self.assertTrue(event.is_set(), "Event has been cleared")
        else:
            self.fail("Exception not raised")

        # Clear
        event.clear()
        self.assertFalse(event.is_set(), "Event has been set")
        self.assertIsNone(event.data, "Non-None data")
        self.assertIsNone(event.exception, "Non-None exception")

    def testWait(self):
        """
        Tests the wait() method
        """
        event = utilities.EventData()
        self.assertFalse(event.wait(.1), "Timed out wait must return False")

        start = time.time()
        threading.Timer(1, event.set).start()
        self.assertFalse(event.wait(.1), "Timed out wait must return False")
        self.assertTrue(event.wait(), "Wait should return true on set")
        self.assertLessEqual(time.time() - start, 2, "Too long to wait")

        self.assertTrue(event.wait(.5),
                        "Already set event shoudn't block wait()")
        self.assertTrue(event.wait(),
                        "Already set event shoudn't block wait()")

    def testWaitException(self):
        """
        Tests the exception effect on wait()
        """
        event = utilities.EventData()
        exception = Exception("Some dummy exception")

        # "Raise" an exception
        threading.Timer(.5, event.raise_exception, [exception]).start()

        # Check the behavior of "wait"
        try:
            event.wait()
        except Exception as ex:
            self.assertIs(ex, exception, "Not the same exception")
        else:
            self.fail("Exception not raised")

        # Check content
        self.assertTrue(event.is_set(), "Event has been cleared")
        self.assertIsNone(event.data, "Non-None data")
        self.assertIs(event.exception, exception, "Invalid exception")

# ------------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Pelix remote services: Common dispatcher

Calls services according to the given method name and parameters

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import json
import logging
import threading

try:
    # Python 3
    # pylint: disable=F0401,E0611
    from urllib.parse import urljoin
    import http.client as httplib
except ImportError:
    # Python 2 or IronPython
    # pylint: disable=F0401
    from urlparse import urljoin
    import httplib

# iPOPO decorators
from pelix.ipopo.decorators import (
    ComponentFactory,
    Requires,
    Provides,
    BindField,
    Property,
    Validate,
    Invalidate,
    Instantiate,
    UnbindField,
)
from pelix.utilities import to_str

# Pelix constants
import pelix.constants
import pelix.framework
import pelix.http
import pelix.remote
import pelix.remote.beans as beans

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------


@ComponentFactory("pelix-remote-dispatcher-factory")
@Provides(pelix.remote.SERVICE_DISPATCHER)
@Requires("_exporters", pelix.remote.SERVICE_EXPORT_PROVIDER, True, True)
@Requires(
    "_listeners",
    pelix.remote.SERVICE_EXPORT_ENDPOINT_LISTENER,
    aggregate=True,
    optional=True,
)
@Instantiate("pelix-remote-dispatcher")
class Dispatcher(object):
    """
    Common dispatcher for all exporters
    """

    def __init__(self):
        # Remote Service providers
        self._exporters = []

        # Injected listeners
        self._listeners = []

        # Bundle context
        self._context = None

        # Framework UID
        self._fw_uid = None

        # Kind -> {Name -> Endpoint}
        self.__kind_endpoints = {}

        # UID -> Endpoint
        self.__endpoints = {}
        self.__endpoints_lock = threading.Lock()

        # UID -> Exporter
        self.__uid_exporter = {}

        # Service Reference -> set(UID)
        self.__service_uids = {}
        self.__exporters_lock = threading.Lock()

    @Validate
    def _validate(self, context):
        """
        Component validated
        """
        # Get the framework UID
        self._fw_uid = context.get_property(pelix.constants.FRAMEWORK_UID)
        self._context = context

        # Prepare the export LDAP filter
        ldapfilter = "(|({0}=*)({1}=*))".format(
            pelix.remote.PROP_EXPORTED_CONFIGS,
            pelix.remote.PROP_EXPORTED_INTERFACES,
        )

        # Export existing services
        existing_ref = context.get_all_service_references(None, ldapfilter)
        if existing_ref is not None:
            for reference in existing_ref:
                self.__export_service(reference)

        # Register a service listener, to update the exported services state
        context.add_service_listener(self, ldapfilter)

    @Invalidate
    def _invalidate(self, context):
        """
        Component invalidated: clean up storage
        """
        # Unregister the service listener
        context.remove_service_listener(self)
        self._context = None
        self._fw_uid = None

    @staticmethod
    def _compute_endpoint_name(properties):
        """
        Computes the end point name according to service properties

        :param properties: Service properties
        :return: The computed end point name
        """
        name = properties.get(pelix.remote.PROP_ENDPOINT_NAME)
        if not name:
            name = "service_{0}".format(properties[pelix.constants.SERVICE_ID])

        return name

    def _check_name_reuse(self, name):
        """
        Checks if a service was waiting to reuse an endpoint name

        :param name: The endpoint name to reuse
        """
        ldap_filter = "(&({0}={1})(!({2}=*)))".format(
            pelix.remote.PROP_ENDPOINT_NAME, name, pelix.remote.PROP_IMPORTED
        )
        svc_ref = self._context.get_service_reference(None, ldap_filter)
        if svc_ref is not None:
            # A service wants to be exported with the given endpoint name
            _logger.debug("Reuse endpoint name %s with %s", name, svc_ref)
            self.__export_service(svc_ref)

    def service_changed(self, event):
        """
        Called when a service event is triggered
        """
        kind = event.get_kind()
        svc_ref = event.get_service_reference()

        with self.__exporters_lock:
            if kind == pelix.framework.ServiceEvent.REGISTERED:
                # Simply export the service
                self.__export_service(svc_ref)

            elif kind == pelix.framework.ServiceEvent.MODIFIED:
                # Matching registering or updated service
                if svc_ref not in self.__service_uids:
                    # New match
                    self.__export_service(svc_ref)

                else:
                    # Properties modification:
                    # Re-export if endpoint.name has changed
                    self.__update_service(
                        svc_ref, event.get_previous_properties()
                    )

            elif svc_ref in self.__service_uids and (
                kind == pelix.framework.ServiceEvent.UNREGISTERING
                or kind == pelix.framework.ServiceEvent.MODIFIED_ENDMATCH
            ):
                # Service is updated or unregistering
                self.__unexport_service(svc_ref)

    def __export_service(self, svc_ref):
        """
        Exports the given service using all available matching providers

        :param svc_ref: A service reference
        """
        if not beans.compute_exported_specifications(svc_ref):
            # No exported specification matches the service
            # (case of iPOPO components services sharing export properties)
            return

        # Service can be exported
        service_uids = self.__service_uids.setdefault(svc_ref, set())

        if not self._exporters:
            _logger.warning("No exporters yet.")
            return

        # Select providers according to the supported configurations
        configs = svc_ref.get_property(pelix.remote.PROP_EXPORTED_CONFIGS)
        if not configs or configs == "*":
            # Export with all providers
            exporters = self._exporters[:]

        else:
            # Filter exporters
            exporters = [
                exporter
                for exporter in self._exporters[:]
                if exporter.handles(configs)
            ]

        if not exporters:
            _logger.warning("No exporter for %s", configs)
            return

        # Prepare an endpoint name
        name = self._compute_endpoint_name(svc_ref.get_properties())

        # Create endpoints
        endpoints = []
        for exporter in exporters:
            try:
                # Create the endpoint
                endpoint = exporter.export_service(svc_ref, name, self._fw_uid)
                if endpoint is None:
                    # Export refused
                    continue

                endpoints.append(endpoint)

                # Store it
                uid = endpoint.uid
                self.__endpoints[uid] = endpoint
                self.__uid_exporter[uid] = exporter
                service_uids.add(uid)

            except (NameError, pelix.constants.BundleException) as ex:
                _logger.error("Error exporting service: %s", ex)

        if not endpoints:
            _logger.warning("No endpoint created for %s", svc_ref)
            return

        # Call listeners (out of the lock)
        if self._listeners:
            for listener in self._listeners[:]:
                listener.endpoints_added(endpoints)

    def __update_service(self, svc_ref, old_properties):
        """
        Service updated, notify exporters
        """
        try:
            # Get the UIDs of its endpoints
            uids = self.__service_uids[svc_ref].copy()

        except KeyError:
            # No known UID
            return

        names = set()
        for uid in uids:
            try:
                # Get its exporter and bean
                exporter = self.__uid_exporter[uid]
                endpoint = self.__endpoints[uid]

            except KeyError:
                # No exporter
                _logger.warning("No exporter for endpoint %s", uid)

                # Remove the UID from the storage
                self.__service_uids[svc_ref].retain(uid)

            else:
                # TODO: check configuration change (can be an unexport)

                # Compute the previous name
                new_name = self._compute_endpoint_name(svc_ref.get_properties())

                try:
                    exporter.update_export(endpoint, new_name, old_properties)

                except NameError as ex:
                    _logger.error("Error updating service properties: %s", ex)

                    # Unexport the service from this exporter
                    del self.__uid_exporter[endpoint.uid]
                    del self.__endpoints[endpoint.uid]
                    exporter.unexport_service(endpoint)
                    names.add(endpoint.name)

                    # Call listeners (out of the lock)
                    if self._listeners:
                        for listener in self._listeners:
                            listener.endpoint_removed(endpoint)

                else:
                    # Call listeners (out of the lock)
                    if self._listeners:
                        for listener in self._listeners:
                            listener.endpoint_updated(endpoint, old_properties)

        # Check if a service wanted to export an endpoint with the given name
        for name in names:
            self._check_name_reuse(name)

    def __unexport_service(self, svc_ref):
        """
        Deletes all endpoints for the given service

        :param svc_ref: A service reference
        """
        try:
            # Get the UIDs of its endpoints
            uids = self.__service_uids.pop(svc_ref)

        except KeyError:
            # No known UID
            return

        names = set()
        for uid in uids:
            try:
                # Remove from storage
                endpoint = self.__endpoints.pop(uid)
                exporter = self.__uid_exporter.pop(uid)

            except KeyError:
                # Oops
                _logger.warning("Trying to remove a lost endpoint (%s)", uid)

            else:
                # Delete endpoint
                exporter.unexport_service(endpoint)
                names.add(endpoint.name)

                # Call listeners
                if self._listeners:
                    for listener in self._listeners[:]:
                        try:
                            listener.endpoint_removed(endpoint)

                        except Exception as ex:
                            _logger.error("Error notifying listener: %s", ex)

        # Check if a service wanted to export an endpoint with the given name
        for name in names:
            self._check_name_reuse(name)

    @BindField("_listeners")
    def _bind_listener(self, field, listener, svc_ref):
        # pylint: disable=W0613
        """
        Listener bound to the component
        """
        # Exported services listener
        if self.__endpoints:
            try:
                listener.endpoints_added(list(self.__endpoints.values()))

            except Exception as ex:
                _logger.exception(
                    "Error notifying newly bound listener: %s", ex
                )

    @BindField("_exporters", if_valid=True)
    def _bind_exporter(self, field, exporter, exporter_ref):
        # pylint: disable=W0613
        """
        Exporter bound
        """
        with self.__exporters_lock:
            # Tell the exporter to export already known services
            for svc_ref in self.__service_uids:
                # Compute the endpoint name
                name = self._compute_endpoint_name(svc_ref.get_properties())

                try:
                    # Create the endpoint
                    endpoint = exporter.export_service(
                        svc_ref, name, self._fw_uid
                    )

                    # Store it
                    uid = endpoint.uid
                    self.__endpoints[uid] = endpoint
                    self.__uid_exporter[uid] = exporter
                    self.__service_uids.setdefault(svc_ref, set()).add(uid)

                except (NameError, pelix.constants.BundleException) as ex:
                    _logger.error("Error exporting service: %s", ex)

                else:
                    # Call listeners (out of the lock)
                    if self._listeners:
                        for listener in self._listeners[:]:
                            listener.endpoints_added([endpoint])

    @UnbindField("_exporters")
    def _unbind_exporter(self, field, exporter, svc_ref):
        # pylint: disable=W0613
        """
        Exporter gone
        """
        removed_endpoints = []

        with self.__exporters_lock:
            # Get the UIDs of all endpoints managed by this exporter
            uids = [
                uid
                for uid, uid_exporter in self.__uid_exporter.items()
                if uid_exporter is exporter
            ]

            # Delete each endpoint
            for uid in uids:
                # Remove references
                del self.__uid_exporter[uid]
                endpoint = self.__endpoints.pop(uid)
                self.__service_uids.get(endpoint.reference, set()).remove(uid)

                # Store the endpoint for listeners
                removed_endpoints.append(endpoint)

                # Unexport the service
                try:
                    exporter.unexport_service(endpoint)

                except Exception as ex:
                    _logger.exception("Error unexporting service: %s", ex)

        # Notify listeners (out of the lock)
        if self._listeners:
            for listener in self._listeners[:]:
                for endpoint in removed_endpoints:
                    try:
                        listener.endpoint_removed(endpoint)
                    except Exception as ex:
                        _logger.error("Error notifying listener: %s", ex)

    def get_endpoint(self, uid):
        """
        Retrieves an end point description, selected by its UID.
        Returns None if the UID is unknown.

        :param uid: UID of an end point
        :return: An :class:`~pelix.remote.beans.ExportEndpoint` or None.
        """
        return self.__endpoints.get(uid)

    def get_endpoints(self, kind=None, name=None):
        """
        Retrieves all end points matching the given kind and/or name

        :param kind: A kind of end point
        :param name: The name of the end point
        :return: A list of :class:`~pelix.remote.beans.ExportEndpoint`
                 matching the parameters
        """
        with self.__endpoints_lock:
            # Get all endpoints
            endpoints = list(self.__endpoints.values())

        # Filter by name
        if name:
            endpoints = [
                endpoint for endpoint in endpoints if endpoint.name == name
            ]

        # Filter by kind
        if kind:
            endpoints = [
                endpoint
                for endpoint in endpoints
                if kind in endpoint.configurations
            ]

        return endpoints


# -----------------------------------------------------------------------------


@ComponentFactory(pelix.remote.FACTORY_REGISTRY_SERVLET)
@Provides(pelix.http.HTTP_SERVLET)
@Provides(pelix.remote.SERVICE_DISPATCHER_SERVLET, "_controller")
@Requires("_dispatcher", pelix.remote.SERVICE_DISPATCHER)
@Requires("_registry", pelix.remote.SERVICE_REGISTRY)
@Property("_path", pelix.http.HTTP_SERVLET_PATH, "/pelix-dispatcher")
class RegistryServlet(object):
    """
    Servlet to access the content of the registry
    """

    def __init__(self):
        """
        Sets up members
        """
        # The framework UID
        self._fw_uid = None

        # The dispatcher
        self._dispatcher = None

        # The imported services registry
        self._registry = None

        # Controller for the provided service:
        # => activate only if bound to a server
        self._controller = False

        # Servlet path property
        self._path = None

        # Ports of exposing servers
        self._ports = []

    @Validate
    def _validate(self, context):
        """
        Component validated
        """
        # Get the framework UID
        self._fw_uid = context.get_property(pelix.constants.FRAMEWORK_UID)

        # Normalize the path
        self._path = "/{0}/".format(
            "/".join(part for part in self._path.split("/") if part)
        )

        _logger.debug(
            "Dispatcher servlet for %s on %s", self._fw_uid, self._path
        )

    @Invalidate
    def _invalidate(self, _):
        """
        Component invalidated
        """
        # Clean up
        self._fw_uid = None

    @staticmethod
    def __grab_data(host, port, path):
        """
        Sends a HTTP request to the server at (host, port), on the given path.
        Returns the parsed response.
        Returns None if the HTTP result is not 200 or in case of error.

        :param host: Dispatcher host address
        :param port: Dispatcher HTTP service port
        :param path: Request path
        :return: The parsed response content, or None
        """
        # Request the end points
        try:
            conn = httplib.HTTPConnection(host, port)
            conn.request("GET", path)
            result = conn.getresponse()
            data = result.read()
            conn.close()
        except Exception as ex:
            _logger.error("Error accessing the dispatcher servlet: %s", ex)
            return None

        if result.status != 200:
            # Not a valid result
            return None

        try:
            # Convert the response to a string
            data = to_str(data)
            # Parse the JSON result
            return json.loads(data)
        except ValueError as ex:
            # Error parsing data
            _logger.error(
                "Error reading the response of the dispatcher: %s", ex
            )

    def _make_endpoint_dict(self, endpoint):
        """
        Converts the end point into a dictionary

        :param endpoint: The end point to convert
        :return: A dictionary
        """
        # Send import-side properties
        return {
            "sender": self._fw_uid,
            "uid": endpoint.uid,
            "configurations": endpoint.configurations,
            "name": endpoint.name,
            "specifications": endpoint.specifications,
            "properties": endpoint.make_import_properties(),
        }

    @staticmethod
    def _make_endpoint_bean(endpoint_dict, host=None):
        """
        Converts an endpoint dictionary into an ImportEndpoint bean

        :param endpoint_dict: Dictionary form of the endpoint
        :param host: The host of the endpoint (optional)
        :return: An ImportEndpoint bean
        """
        # Create the end point bean
        endpoint = pelix.remote.beans.ImportEndpoint(
            endpoint_dict["uid"],
            endpoint_dict["sender"],
            endpoint_dict["configurations"],
            endpoint_dict["name"],
            endpoint_dict["specifications"],
            endpoint_dict["properties"],
        )

        # Set the host address
        endpoint.server = host
        return endpoint

    def bound_to(self, path, parameters):
        # pylint: disable=W0613
        """
        This servlet has been bound to a server

        :param path: The servlet path in the server
        :param parameters: The servlet/server parameters
        """
        port = parameters["http.port"]
        if port not in self._ports:
            # Get its access port
            self._ports.append(port)

            # Activate the service, we're bound to a server
            self._controller = True

    def unbound_from(self, path, parameters):
        # pylint: disable=W0613
        """
        This servlet has been unbound from a server

        :param path: The servlet path in the server
        :param parameters: The servlet/server parameters
        """
        port = parameters["http.port"]
        if port in self._ports:
            # Remove its access port
            self._ports.remove(port)

            # Deactivate the service if no more server available
            if not self._ports:
                self._controller = False

    def do_GET(self, request, response):
        """
        Handles a GET request

        :param request: Request handler
        :param response: Response handler
        """
        # Normalize the path
        path_parts = [part for part in request.get_path().split("/") if part]

        # Remove the servlet part
        servlet_parts = [part for part in self._path.split("/") if part]
        path_parts = path_parts[len(servlet_parts) :]
        action = path_parts[0]

        if action == "framework":
            # /framework: return the framework UID, let it be converted as a
            # JSON string
            data = self._fw_uid
        elif action == "endpoints":
            # /endpoints: all end points
            endpoints = self._dispatcher.get_endpoints()
            if not endpoints:
                data = []
            else:
                data = [
                    self._make_endpoint_dict(endpoint) for endpoint in endpoints
                ]
        elif action == "endpoint":
            # /endpoint/<uid>: specific end point
            try:
                uid = path_parts[1]
                endpoint = self._dispatcher.get_endpoint(uid)
            except IndexError:
                # UID not given
                uid = "<unknown>"
                endpoint = None

            if endpoint is None:
                response.send_content(
                    404, "Unknown UID: {0}".format(uid), "text/plain"
                )
                return
            else:
                data = self._make_endpoint_dict(endpoint)
        else:
            # Unknown
            response.send_content(
                404,
                "Unhandled path {0}".format(request.get_path()),
                "text/plain",
            )
            return

        # Convert the result to JSON
        data = json.dumps(data)

        # Send the result
        response.send_content(200, data, "application/json")

    def do_POST(self, request, response):
        """
        Handles a POST request

        :param request: Request handler
        :param response: Response handler
        """
        # Split the path
        path_parts = request.get_path().split("/")
        if path_parts[-1] != "endpoints":
            # Bad path
            response.send_content(404, "Unhandled path", "text/plain")
            return

        # Read the content
        endpoints = json.loads(to_str(request.read_data()))
        if endpoints:
            # Got something
            sender = request.get_client_address()[0]
            for endpoint in endpoints:
                self._registry.add(self._make_endpoint_bean(endpoint, sender))

        # We got the end points
        response.send_content(200, "OK", "text/plain")

    def get_access(self):
        """
        Returns the port and path to access this servlet with the first
        bound HTTP service.
        Returns None if this servlet is still not bound to a server

        :return: A tuple: (port, path) or None
        """
        if self._ports:
            return self._ports[0], self._path

        return None

    def grab_endpoint(self, host, port, path, uid):
        """
        Retrieves the description of the end point with the given UID from the
        given dispatcher servlet.
        Returns an ImportEndpoint bean, or None in case of error.
        Does not register the end point.

        :param host: Dispatcher host address
        :param port: Dispatcher HTTP service port
        :param path: Path to the dispatcher servlet
        :param uid: The UID of an end point
        :return: An ImportEndpoint bean or None
        """
        # Setup the request URI
        if path[-1] == "/":
            path = path[:-1]

        request_path = "{0}/endpoint/{1}".format(path, uid)

        # Get the endpoint description
        endpoint_dict = self.__grab_data(host, port, request_path)
        if not endpoint_dict:
            # No description found
            return None

        # Create the end point bean
        return self._make_endpoint_bean(endpoint_dict, host)

    def send_discovered(self, host, port, path):
        """
        Sends a "discovered" HTTP POST request to the dispatcher servlet of the
        framework that has been discovered

        :param host: The address of the sender
        :param port: Port of the HTTP server of the sender
        :param path: Path of the dispatcher servlet
        :return: True if the request has been handled by the peer
        """
        # Get the end points from the dispatcher
        endpoints = [
            self._make_endpoint_dict(endpoint)
            for endpoint in self._dispatcher.get_endpoints()
        ]

        # Make the path to /endpoints
        if path[-1] != "/":
            path += "/"
        path = urljoin(path, "endpoints")

        # Request the end points
        try:
            conn = httplib.HTTPConnection(host, port)
            conn.request(
                "POST",
                path,
                json.dumps(endpoints),
                {"Content-Type": "application/json"},
            )
            result = conn.getresponse()
            data = result.read()
            conn.close()
        except Exception as ex:
            _logger.error(
                "Error sending endpoints to the framework at %s:%s: %s",
                host,
                port,
                ex,
            )
            return False
        else:
            if result.status != 200:
                # Not a valid result
                _logger.warning(
                    "Got an HTTP code %d when contacting a "
                    "discovered framework: %s",
                    result.status,
                    data,
                )
            return result.status == 200



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
FileInstall: Polls for changes on files in a directory and notifies listeners

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import logging
import os
import threading
import zlib

# Pelix
from pelix.ipopo.decorators import (
    ComponentFactory,
    Provides,
    Requires,
    Validate,
    Invalidate,
    Instantiate,
    BindField,
    UnbindField,
    UpdateField,
    Property,
)
import pelix.services as services
import pelix.threadpool

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------


@ComponentFactory()
@Provides(services.SERVICE_FILEINSTALL)
@Requires(
    "_listeners",
    services.SERVICE_FILEINSTALL_LISTENERS,
    aggregate=True,
    optional=True,
)
@Property("_poll_time", "poll.time", 1)
@Instantiate("pelix-services-file-install")
class FileInstall(object):
    """
    Polls folders to look for files modifications
    """

    def __init__(self):
        """
        Sets up members
        """
        # Listeners (injected)
        self._listeners = []

        # Folder -> [listeners] (computed)
        self._folder_listeners = {}

        # Polling delta time (1 second by default)
        self._poll_time = 1

        # Lock
        self.__lock = threading.RLock()

        # Single thread task pool to notify listeners
        self.__pool = pelix.threadpool.ThreadPool(
            1, logname="FileInstallNotifier"
        )

        # 1 thread per watched folder (folder -> Thread)
        self.__threads = {}

        # Thread stoppers (folder -> Event)
        self.__stoppers = {}

        # Validation flag
        self.__validated = False

    @Validate
    def _validate(self, _):
        """
        Component validated
        """
        with self.__lock:
            # Start the task pool
            self.__pool.start()

            # Update the flag
            self.__validated = True

    @Invalidate
    def _invalidate(self, _):
        """
        Component invalidated
        """
        with self.__lock:
            # Update the flag
            self.__validated = False

            # Stop all threads
            for event in set(self.__stoppers.values()):
                event.set()

            # Wait for them
            for thread in set(self.__threads.values()):
                thread.join()

            # Stop the task pool
            self.__pool.stop()

            # Clean up
            self.__stoppers.clear()
            self.__threads.clear()

    @BindField("_listeners")
    def _bind_listener(self, _, svc, svc_ref):
        """
        A new listener is bound
        """
        with self.__lock:
            folder = svc_ref.get_property(services.PROP_FILEINSTALL_FOLDER)
            if folder:
                # Register the listener for this service
                self.add_listener(folder, svc)

    @UpdateField("_listeners")
    def _update_field(self, _, svc, svc_ref, old_props):
        """
        A bound listener has been updated
        """
        with self.__lock:
            old_folder = old_props.get(services.PROP_FILEINSTALL_FOLDER)
            new_folder = svc_ref.get_property(services.PROP_FILEINSTALL_FOLDER)

            if old_folder != new_folder:
                # Folder changed
                self.remove_listener(old_folder, svc)
                self.add_listener(new_folder, svc)

    @UnbindField("_listeners")
    def _unbind_listener(self, _, svc, svc_ref):
        """
        A listener is gone
        """
        with self.__lock:
            folder = svc_ref.get_property(services.PROP_FILEINSTALL_FOLDER)
            if folder:
                # Remove the listener
                self.remove_listener(folder, svc)

    def add_listener(self, folder, listener):
        """
        Manual registration of a folder listener

        :param folder: Path to the folder to watch
        :param listener: Listener to register
        :return: True if the listener has been registered
        """
        with self.__lock:
            # Simply add the listener
            if folder:
                try:
                    listeners = self._folder_listeners[folder]

                except KeyError:
                    # Unknown folder
                    listeners = self._folder_listeners[folder] = set()

                    # Start a new thread
                    event = self.__stoppers[folder] = threading.Event()
                    thread = threading.Thread(
                        target=self.__watch,
                        args=(folder, event),
                        name="FileInstall-{0}".format(folder),
                    )
                    thread.daemon = True
                    self.__threads[folder] = thread
                    thread.start()

                listeners.add(listener)
                return True

            return False

    def remove_listener(self, folder, listener):
        """
        Manual unregistration of a folder listener.

        :param folder: Path to the folder the listener watched
        :param listener: Listener to unregister
        :raise ValueError: The listener wasn't watching this folder
        """
        with self.__lock:
            # Remove the listener
            listeners = self._folder_listeners[folder]
            listeners.remove(listener)
            if not listeners:
                try:
                    # Stop the corresponding thread
                    self.__stoppers.pop(folder).set()

                except KeyError:
                    # Component invalidated
                    pass

                else:
                    # Wait for the thread to stop
                    self.__threads.pop(folder).join()

                    # No more listener for this folder
                    del self._folder_listeners[folder]

    def __notify(self, folder, added, updated, deleted):
        """
        Notifies listeners that files of a folder has been modified

        :param folder: Folder where changes occurred
        :param added: Names of added files
        :param updated: Names of modified files
        :param deleted: Names of removed files
        """
        with self.__lock:
            try:
                # Get a copy of the listeners for this folder
                listeners = self._folder_listeners[folder].copy()

            except KeyError:
                # No (more) listeners: do nothing
                return

        for listener in listeners:
            try:
                listener.folder_change(folder, added, updated, deleted)

            except Exception as ex:
                _logger.exception("Error notifying a folder listener: %s", ex)

    @staticmethod
    def __get_checksum(filepath):
        """
        Returns the checksum (Adler32) of the given file

        :param filepath: Path to the file
        :return: The checksum (int) of the given file
        :raise OSError: File not accessible
        :raise IOError: File not readable
        """
        # Don't forget to open the file in binary mode
        with open(filepath, "rb") as filep:
            # Return the checksum of the given file
            return zlib.adler32(filep.read())

    def __get_file_info(self, folder, filename):
        """
        Returns the (mtime, checksum) tuple for the given file

        :param folder: Path to the parent folder
        :param filename: Base name of the file
        :return: A tuple containing file information
        :raise OSError: File not accessible
        :raise IOError: File not readable
        """
        filepath = os.path.join(folder, filename)
        return os.path.getmtime(filepath), self.__get_checksum(filepath)

    def __check_different(self, folder, filename, file_info, updated):
        """
        Checks if the given file has changed since the previous check

        :param folder: Path to the parent folder
        :param filename: Base name of the file
        :param file_info: Current information about the file
        :param updated: Set of updated files, where the file name might be
                        added
        :return: The (updated) file information tuple
        :raise OSError: File not accessible
        :raise IOError: File not readable
        """
        # Compute the file path
        filepath = os.path.join(folder, filename)

        # Get the previous modification time
        previous_mtime = file_info[0]

        # Get the new modification time
        mtime = os.path.getmtime(filepath)

        if previous_mtime == mtime:
            # No modification (no need to compute the checksum)
            return file_info

        # Get the previous checksum
        previous_checksum = file_info[1]

        # Compute the new one
        checksum = self.__get_checksum(filepath)

        if previous_checksum == checksum:
            # No real modification, update file info
            return mtime, checksum

        # File modified
        updated.add(filename)
        return mtime, checksum

    def __watch(self, folder, stopper):
        """
        Loop that looks for changes in the given folder

        :param folder: Folder to watch
        :param stopper: An Event object that will stop the loop once set
        """
        # File name -> (modification time, checksum)
        previous_info = {}

        while not stopper.wait(self._poll_time) and not stopper.is_set():
            if not os.path.exists(folder):
                # Nothing to do yet
                continue

            # Look for files
            filenames = {
                filename
                for filename in os.listdir(folder)
                if os.path.isfile(os.path.join(folder, filename))
            }

            # Prepare the sets
            added = set()
            updated = set()
            deleted = set(previous_info.keys()).difference(filenames)

            # Compute differences
            for filename in filenames:
                try:
                    # Get previous information
                    file_info = previous_info[filename]

                except KeyError:
                    # Unknown file: added one
                    added.add(filename)
                    previous_info[filename] = self.__get_file_info(
                        folder, filename
                    )

                else:
                    try:
                        # Known file name
                        new_info = self.__check_different(
                            folder, filename, file_info, updated
                        )
                        # Store new information
                        previous_info[filename] = new_info

                    except (IOError, OSError):
                        # Error reading file, do nothing
                        pass

            # Remove information about deleted files
            for filename in deleted:
                del previous_info[filename]

            if added or updated or deleted:
                # Something changed: notify listeners
                self.__pool.enqueue(
                    self.__notify, folder, added, updated, deleted
                )
